/*   
     
Chess-Ec    Copyright (C)   2001-2008 Marc Dechico


 ec_branche.c: Memorize the subsequants move used to get the best move.
   
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 To contact the author:	Marc Dechico
 Kerfring
 29390 SCAER France
 tel: 06 37 98 22 54
 Mail: ec@chess-ec.org  
    
*/
/* target of the table:
   a linked liste of move (like E2E4->E7E5->D2D4......)  is created along
   when uploading the best note. 
   to avoid allocating memory each time the best move is selected,
   memory chunq are automatically created   
   
   each node(a move)  is memorized and linked with this structure:
   struct  st_bra_nod
   {
      char c_pos_n;                 	  ariving square
      char c_typ_mvt;		          move type  
      char c_pos_ori_n;                    departure square 
      struct branch_node *pst_nxt_nod;    next _subsequent node (move) 
   };

   as a first evalulation upload branche cost 3.1 precent of simulation time 
*/
#include "ec_branche.h"

#ifdef UPLOAD_BRANCHE


int bra_ini_mem (long li_ini_nod_qua,int i_inc_per)
{
  /* initialise memory for the branche bloc */

  st_branche.li_ini_nod_qua=li_ini_nod_qua;
  st_branche.li_tot_nod_nbr=li_ini_nod_qua;
  st_branche.i_inc_per=i_inc_per;

  /* memorize quantitiy of memory used */
  st_branche.li_byt_siz =sizeof(struct st_bra_nod) *  st_branche.li_ini_nod_qua;

  /* no increment yet */
  st_branche.i_tot_inc_nbr=0;
 
  /* alloc memory */
  st_branche.pst_blo_beg=mem_malloc(st_branche.li_byt_siz); 
  st_branche.pst_fre_new_nod= st_branche.pst_blo_beg;
  st_branche.pst_nxt_fre_nod= st_branche.pst_blo_beg;

  /* no garbage for the moment */
  st_branche.pst_fre_old_nod= NULL;
  st_branche.pst_las_old_nod= NULL;

  /* memorise next fee node */
  st_branche.pst_fre_new_nod= st_branche.pst_nxt_fre_nod=st_branche.pst_blo_beg;

  /* memorise end-adresse of memory block*/
  st_branche.pst_blo_end=st_branche.pst_blo_beg -2 + st_branche.li_ini_nod_qua;
 
  /* initialize memory */ 
  memset(st_branche.pst_blo_beg,0,st_branche.li_byt_siz);

  /*initialize garbage collector */
   mem_lst_ini(&st_branche.pst_ind);

  /* add this memory-block to the garbage collector */
  mem_lst_add(&st_branche.pst_ind,(void *) st_branche.pst_blo_beg);
  /* debug */
  //printf("\n init memory ba  pst_ind=%li pre_ind=%li",(long)st_branche.pst_ind,(long)st_branche.pst_ind->pst_pre_ind);
  return 0;
}
int bra_fre_mem              (void)
{
  mem_lst_fre(&st_branche.pst_ind);

  return 0;
}

static int bra_inc_mem(void)
{
  //printf("\n Incremente bloc branche"); 
  /* increment  memory for the branche table */

  
  /* quantitiy of nodes for  increment */
  long i_nbr_nod_inc   =  st_branche.li_ini_nod_qua  * st_branche.i_inc_per / 100.00;
  long li_byt_siz_inc  = sizeof(struct st_bra_nod) * i_nbr_nod_inc;
  
  st_branche.li_byt_siz+=li_byt_siz_inc;

  /* memorize how many increment  */
  st_branche.i_tot_inc_nbr++;
 
  /* alloc memory */
  st_branche.pst_fre_new_nod=  st_branche.pst_nxt_fre_nod=mem_malloc(li_byt_siz_inc); 

  /* memorise end-adresse of memory block*/
  st_branche.pst_blo_end=st_branche.pst_fre_new_nod -2 +i_nbr_nod_inc ;
 
  /* initialize memory */ 
  memset(st_branche.pst_fre_new_nod,0,li_byt_siz_inc);

  /* add this memory-block to the garbage collector */
  mem_lst_add(&st_branche.pst_ind,(void *)st_branche.pst_fre_new_nod);

  /* debug */
  //  printf("\nadd increment  nodr pst_ind=%li pre_ind=%li",(long)st_branche.pst_ind,(long)st_branche.pst_ind->pst_pre_ind);
  return 0;



}
int bra_get_nxt_nod (void) 
{
  if(st_branche.pst_fre_old_nod!=NULL)
    {
      st_branche.pst_nxt_fre_nod=st_branche.pst_fre_old_nod;
      st_branche.pst_fre_old_nod= st_branche.pst_nxt_fre_nod->pst_nxt_nod;
    }
  else
    {

      st_branche.li_tot_nod_nbr++;

      /* verify if more memory has to be allocated */
      if (st_branche.pst_blo_end == st_branche.pst_fre_new_nod)
	{
	  bra_inc_mem();
	}
      else
	{
	  /* increment next free segment pointer  for arr */
	  st_branche.pst_fre_new_nod ++;
	  st_branche.pst_nxt_fre_nod=st_branche.pst_fre_new_nod;
	}
    }
  return 0;
}


int bra_add_nod_max_inf(struct st_bra_nod **ppst_bra_nod,
			struct st_bra_nod **ppst_bra_nod_las,
			char * c_pos_ori_n, char * c_pos_n, char *c_typ_mvt)
{
  if (*ppst_bra_nod == NULL)
    {
      *ppst_bra_nod=st_branche.pst_nxt_fre_nod;

      /* opti  not making this update and use diffeent function */
      /* in pie_sui depending on the level */
      *ppst_bra_nod_las=st_branche.pst_nxt_fre_nod;

      bra_get_nxt_nod();
    }
  
  (*ppst_bra_nod)->i_pos_ori_n= * c_pos_ori_n;
  (*ppst_bra_nod)->i_pos_n= * c_pos_n;
  (*ppst_bra_nod)->i_typ_mvt= * c_typ_mvt;
  (*ppst_bra_nod)->pst_nxt_nod= NULL; 
  /*
  printf("\nmax_inf:");
  bra_prt_lst(*ppst_bra_nod);
  */
  return 0;
}
int bra_add_nod_lea (struct st_bra_nod **ppst_bra_nod_hea,
		     struct st_bra_nod **ppst_bra_nod_tai,
		     struct st_bra_nod **ppst_bra_nod_las_p,
		     struct st_bra_nod **ppst_bra_nod_las,
		     char * c_pos_ori_n, char * c_pos_n, char *c_typ_mvt)
{
  /* leave (d'ont reuse the node  beacause the list at this level  will be used later */
  /* for printing with the note */
  *ppst_bra_nod_hea=st_branche.pst_nxt_fre_nod;
  bra_get_nxt_nod();
  
/* record the actual best move in the nod */
  (*ppst_bra_nod_hea)->i_pos_ori_n= * c_pos_ori_n;
  (*ppst_bra_nod_hea)->i_pos_n= * c_pos_n;
  (*ppst_bra_nod_hea)->i_typ_mvt= * c_typ_mvt;
  /* add the tail for this best move */
  (*ppst_bra_nod_hea)->pst_nxt_nod= *ppst_bra_nod_tai; 
  /*
  extern int i_pro_sim_act;
  printf("\nadd_nod_lea  pro %i :",i_pro_sim_act);
  bra_prt_lst(*ppst_bra_nod_hea);
  */
  return 0;

}

int bra_add_nod(struct st_bra_nod **ppst_bra_nod_hea,
		struct st_bra_nod **ppst_bra_nod_tai,
		struct st_bra_nod **ppst_bra_nod_las_p,
		struct st_bra_nod **ppst_bra_nod_las,
		char * c_pos_ori_n, char * c_pos_n, char *c_typ_mvt)
{
  /* opti ??? a way to avoid test == NULL ? */
  if (*ppst_bra_nod_hea == NULL )
    {
      /* no best move has yet been found because this is the first move at this level*/
      *ppst_bra_nod_hea=st_branche.pst_nxt_fre_nod;
      bra_get_nxt_nod();
    }
  else
    {
      /* a best move had already been found I am reusing the head of the list */
      /* the tail of the list is hadded to the garbage collector if exist  */  
      if((*ppst_bra_nod_hea)->pst_nxt_nod!=NULL)
	{
	  if(st_branche.pst_las_old_nod!=NULL)
	    st_branche.pst_las_old_nod->pst_nxt_nod=(*ppst_bra_nod_hea)->pst_nxt_nod;
	  else
	    st_branche.pst_las_old_nod=(*ppst_bra_nod_hea)->pst_nxt_nod;

	  /* update the last node of gabage list for future add */
	  st_branche.pst_las_old_nod= *ppst_bra_nod_las_p;
	  
	}
    }
  /* upload also the last nod to previous level */
  *ppst_bra_nod_las_p=*ppst_bra_nod_las;
  /* record the actual best move in the nod */
  (*ppst_bra_nod_hea)->i_pos_ori_n= * c_pos_ori_n;
  (*ppst_bra_nod_hea)->i_pos_n= * c_pos_n;
  (*ppst_bra_nod_hea)->i_typ_mvt= * c_typ_mvt;
  /* add the tail for this best move */
  (*ppst_bra_nod_hea)->pst_nxt_nod= *ppst_bra_nod_tai; 
  /*
  extern int i_pro_sim_act;
  printf("\nadd_nod pro %i :",i_pro_sim_act);
  bra_prt_lst(*ppst_bra_nod_hea);
  */
  return 0;
 
}


int bra_rem_nod(struct st_bra_nod **ppst_bra_nod_hea,
		struct st_bra_nod **ppst_bra_nod_las_p,
		struct st_bra_nod **ppst_bra_nod_las
		)
{
  /* add the removed list  to the garbage */
  if (st_branche.pst_las_old_nod != NULL)
    {
      /* there is allready a node in the garbage */
      /* cat the two list */
      st_branche.pst_las_old_nod->pst_nxt_nod=*ppst_bra_nod_hea;
    }
  else
    {
      /* the garbage is empty */ 
      st_branche.pst_las_old_nod=*ppst_bra_nod_hea;
    }

  /* update the last node of gabage list for future add */
  st_branche.pst_las_old_nod= *ppst_bra_nod_las;
  /* upload also the last nod to previous level */
  *ppst_bra_nod_las_p=*ppst_bra_nod_las;
  /*
  printf("\nbra_rem pro %i :",i_pro_sim_act);
  bra_prt_lst(*ppst_bra_nod_hea);
  */
  return 0;
 
}
int bra_prt_lst(struct st_bra_nod *pst_bra_nod_hea)
{
  for(;;)
    {
      if (pst_bra_nod_hea != NULL)
	{
	  printf("%s",tc_lib_cas[(int)pst_bra_nod_hea->i_pos_ori_n]);
	  printf("%s",tc_lib_cas[(int)pst_bra_nod_hea->i_pos_n]);
	  printf("%c|",pro_pio_tra_tmvt_lpie ((int)pst_bra_nod_hea->i_typ_mvt));

	  if ((pst_bra_nod_hea)->pst_nxt_nod == NULL)
	    break;
	  else
	    {
	      pst_bra_nod_hea=(pst_bra_nod_hea)->pst_nxt_nod;
	    }
	}
    }
  //  getchar();
  return 0;
}
/* ifdef UPLOAD_BRANCHE */
#endif

